''''''''''''''''''''''''''''''''''''''''
' Program    : Loads TIFF Files        '
' Name       : OPENTIFF Version 0.1    '
' Programmer : Yousuf Philips          '
' Company    : Y P I                   '
' Updated On : 12th of June 1999       '
' Email - [philipz85@hotmail.com]      '
' [http://members.xoom.com/Philipz/]   '
''''''''''''''''''''''''''''''''''''''''

'/*               Do not edit this file if you distribute it.              */'
'/*  (c) Copyrighted by YPI in 1998 | All Rights Reserved | Public Domain  */'

'/* This program has been created by YPI (Basic Programming Incorporation) */'
'/* as a utility to load Tagged Images File Format [TIFF] files.           */'

'/* This viewer currently loads 4, 8 and 24 bit uncompressed TIFF files.   */'
'/* It also loads 4, 8 and 24-bit Packbit Compressed TIFF files.           */'
'/* Hopefully in the next version LZW compressed & JPEG compressed TIFF    */'
'/* will be able to loaded.                                                */'

'/* If you use any of this code in your program then you must credit YPI.  */'
'/* YPI has created an assortment of graphics viewers and players and they */'
'/* can all be found at <http://members.xoom.com/Philipz/>                 */'

DECLARE SUB OpenTIFF (FileName$)
DECLARE FUNCTION Sign2Unsign& (Integerx%)
DECLARE FUNCTION Unsign2Sign% (Integerx&)
DECLARE FUNCTION GetBytesAtOffset$ (Offset&, FileNumber%, NumberOfBytes&)

TYPE TIFFHeader
   ByteOrder AS STRING * 2        ' Byte Order 'II' - Intel or 'MM' - Motorola
   Version AS INTEGER             ' ?? Version Number (should be 42)
   OffsetOfIFD AS LONG            ' Offset of First Image File Directory
END TYPE

TYPE TIFFIFD
   NumOfDirEntries AS INTEGER     ' Number of Directory Entries
END TYPE

TYPE TIFFFDE                      ' File Directory Entry
   Tag AS INTEGER                 ' Tag identifier
   DataType AS INTEGER            ' Data Type in Entry
   DataTypeCount AS LONG          ' Number of Data Items (not bytes)
   OffsetOfData AS LONG           ' Offset of Data
END TYPE

'/* The information from the TIFF file is collected and put into this      */'
'/* location for easy access.                                              */'
TYPE TIFFInfo
   TIFFWidth AS LONG              ' Width of Image
   TIFFHeight AS LONG             ' Height of Image
   TIFFBitsPerPixel AS INTEGER    ' Bits Per Pixel
   TIFFPlanes AS INTEGER          ' Number Of Planes
   TIFFCompression AS INTEGER     ' Compression Used
   TIFFRowsinStrip AS INTEGER     ' Number of Rows In Strips
   TIFFStrips AS INTEGER          ' Number of Strips
   RGBOffset AS LONG              ' Offset of Palette
   RGBSize AS INTEGER             ' Size of Palette
   RGBDataType AS INTEGER         ' Palette Data Type
END TYPE

DEFINT A-Z
'ON ERROR GOTO Errrr

CLS
FILES "*.tif"
PRINT " TIFF Viewer for QBASIC - OPENTIFF Ver 0.1"
PRINT " (c) YPI (BASIC Programming Incorporation) 1999"
PRINT " Please Send Comments & Suggestions To <philipz85@hotmail.com>"
INPUT " Enter File Name - ", FileName$
IF RIGHT$(UCASE$(FileName$), 4) <> ".TIF" THEN
   FileName$ = FileName$ + ".TIF"
END IF
CALL OpenTIFF(FileName$)
SYSTEM

Errrr:
RESUME NEXT

FUNCTION GetBytesAtOffset$ (Offset&, FileNumber, NumberOfBytes&)
                      
location& = LOC(FileNumber) + 1
SEEK #FileNumber, Offset&
Word$ = SPACE$(NumberOfBytes&)
GET #FileNumber, , Word$
SEEK #1, location&
GetBytesAtOffset$ = Word$

END FUNCTION

SUB OpenTIFF (FileName$)

DIM TIFFHeader AS TIFFHeader
DIM ImageFD AS TIFFIFD
DIM DirEntry AS TIFFFDE
DIM TIFFInfo AS TIFFInfo
Word$ = SPACE$(2)
DWord$ = SPACE$(4)

CLS
OPEN FileName$ FOR BINARY AS #1
IF LOF(1) = 0 THEN
   PRINT " File Is Empty"
   CLOSE
   KILL FileName$
   EXIT SUB
END IF

GET #1, , TIFFHeader
IF TIFFHeader.ByteOrder = "II" THEN
ELSEIF TIFFHeader.ByteOrder = "MM" THEN
   PRINT " Please Send File To <philipz85@hotmail.com>"
   EXIT SUB
ELSE
   PRINT " Invalid TIFF File"
   CLOSE
   EXIT SUB
END IF
IF TIFFHeader.Version = 42 THEN
ELSE
   PRINT " No Confirmation That File Is TIFF File"
   EXIT SUB
END IF

SEEK #1, TIFFHeader.OffsetOfIFD + 1
GET #1, , ImageFD

FOR NumberOfEntries = 0 TO ImageFD.NumOfDirEntries - 1
   GET #1, , DirEntry
   SELECT CASE DirEntry.DataType
      CASE 0: 'I don't know what to put here
      CASE 1 TO 2: DataTypeInBytes = 1
      CASE 3: DataTypeInBytes = 2
      CASE 4: DataTypeInBytes = 4
      CASE 5: DataTypeInBytes = 8
      CASE 6 TO 7: DataTypeInBytes = 1
      CASE 8: DataTypeInBytes = 2
      CASE 9: DataTypeInBytes = 4
      CASE 10: DataTypeInBytes = 8
      CASE 11: DataTypeInBytes = 4
      CASE 12: DataTypeInBytes = 8
   END SELECT
   BytesToExtract& = DataTypeInBytes * DirEntry.DataTypeCount
  
   SELECT CASE Sign2Unsign(DirEntry.Tag)
      CASE 256: TIFFInfo.TIFFWidth = DirEntry.OffsetOfData
      CASE 257: TIFFInfo.TIFFHeight = DirEntry.OffsetOfData
      CASE 258:
                IF BytesToExtract& < 5 THEN
                   TIFFInfo.TIFFBitsPerPixel = DirEntry.OffsetOfData
                ELSE
                   Bytez$ = GetBytesAtOffset(DirEntry.OffsetOfData, 1, BytesToExtract&)
                   TIFFInfo.TIFFBitsPerPixel = ASC(MID$(Bytez$, 2, 1))
                END IF
      CASE 259: TIFFInfo.TIFFCompression = Unsign2Sign(DirEntry.OffsetOfData)
      CASE 273:
                IF BytesToExtract& < 5 THEN
                   DIM StripRows(1, 2) AS LONG
                   StripRows(1, 1) = DirEntry.OffsetOfData + 1
                   TIFFInfo.TIFFStrips = 1
                ELSE
                   TIFFInfo.TIFFStrips = (BytesToExtract& \ DataTypeInBytes)
                   DIM StripRows(TIFFInfo.TIFFStrips, 2) AS LONG
                   Word$ = GetBytesAtOffset(DirEntry.OffsetOfData + 1, 1, BytesToExtract&)
                   FOR Loops = 0 TO TIFFInfo.TIFFStrips - 1
                      StripRows(Loops + 1, 1) = CVL(MID$(Word$, (Loops * 4) + 1, 4)) + 1
                   NEXT Loops
                END IF
      CASE 277: TIFFInfo.TIFFPlanes = DirEntry.OffsetOfData
      CASE 278: TIFFInfo.TIFFRowsinStrip = DirEntry.OffsetOfData
      CASE 279:
                IF TIFFInfo.TIFFStrips = 1 THEN
                   StripRows(1, 2) = DirEntry.OffsetOfData
                ELSE
                   Word$ = GetBytesAtOffset(DirEntry.OffsetOfData + 1, 1, BytesToExtract&)
                   FOR Loops = 0 TO TIFFInfo.TIFFStrips - 1
                      StripRows(Loops + 1, 2) = CVL(MID$(Word$, (Loops * 4) + 1, 4))
                   NEXT Loops
                END IF
      CASE 320: TIFFInfo.RGBOffset = DirEntry.OffsetOfData
                TIFFInfo.RGBSize = BytesToExtract&
                TIFFInfo.RGBDataType = DataTypeInBytes
   END SELECT
NEXT

PRINT " Valid TIFF File"
PRINT " Width -"; TIFFInfo.TIFFWidth
PRINT " Height -"; TIFFInfo.TIFFHeight
PRINT " Bits Per Pixel -"; TIFFInfo.TIFFBitsPerPixel * TIFFInfo.TIFFPlanes
PRINT " Compression Used -";
SELECT CASE TIFFInfo.TIFFCompression
   CASE 1: PRINT " Uncompressed"
   CASE 2: PRINT " CCITT 1D": EXIT SUB
   CASE 3 TO 4: PRINT " Group 3 & 4 Fax": EXIT SUB
   CASE 5: PRINT " LZW Compressed": EXIT SUB
   CASE 6: PRINT " JPEG Compressed": EXIT SUB
   CASE -32763: PRINT " Packbit Compressed"
   CASE ELSE: PRINT " Unknown Compression": EXIT SUB
END SELECT
Wait$ = INPUT$(1)

GET #1, , DWord$
IF CVL(DWord$) = 0 THEN
   SCREEN 13
   IF TIFFInfo.RGBOffset <> 0 THEN
      SEEK #1, TIFFInfo.RGBOffset + 1
      PaletteR$ = SPACE$(TIFFInfo.RGBSize \ 3)
      PaletteG$ = SPACE$(TIFFInfo.RGBSize \ 3)
      PaletteB$ = SPACE$(TIFFInfo.RGBSize \ 3)
      GET #1, , PaletteR$
      GET #1, , PaletteG$
      GET #1, , PaletteB$
      OUT &H3C8, 0
      FOR Loops = 0 TO (2 ^ TIFFInfo.TIFFBitsPerPixel) - 1
         OUT &H3C9, ASC(MID$(PaletteR$, Loops * 2 + 1, 1)) \ 4
         OUT &H3C9, ASC(MID$(PaletteG$, Loops * 2 + 1, 1)) \ 4
         OUT &H3C9, ASC(MID$(PaletteB$, Loops * 2 + 1, 1)) \ 4
      NEXT Loops
   ELSE
      '/* For Grayscale Images or 24-bit images */'
      OUT &H3C8, 0
      FOR Loops = 0 TO 255
         OUT &H3C9, Loops \ 4
         OUT &H3C9, Loops \ 4
         OUT &H3C9, Loops \ 4
      NEXT Loops
   END IF
  
   DividerX! = (TIFFInfo.TIFFWidth / 320)
   DividerY! = (TIFFInfo.TIFFHeight / 200)
   DividerX1! = (320 / TIFFInfo.TIFFWidth)
   DividerY1! = (200 / TIFFInfo.TIFFHeight)
   IF DividerX1! > 1 THEN DividerX1! = 1
   IF DividerY1! > 1 THEN DividerY1! = 1

   IF TIFFInfo.TIFFCompression = 1 THEN
      YHeight = 0: BytesToExtractPerStrip$ = SPACE$(TIFFInfo.TIFFWidth * TIFFInfo.TIFFPlanes * (TIFFInfo.TIFFBitsPerPixel / 8))
      FOR StripNo = 1 TO TIFFInfo.TIFFStrips
         SEEK #1, StripRows(StripNo, 1)
         FOR RowNo = 1 TO TIFFInfo.TIFFRowsinStrip
            GET #1, , BytesToExtractPerStrip$: WordMovement = 0
            IF TIFFInfo.TIFFBitsPerPixel = 4 THEN
               FOR XWidth = 0 TO TIFFInfo.TIFFWidth - 1 STEP 2
                  PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(MID$(BytesToExtractPerStrip$, WordMovement + 1, 1)) \ 16
                  PSET ((XWidth * DividerX1!) + 1, YHeight * DividerY1!), ASC(MID$(BytesToExtractPerStrip$, WordMovement + 1, 1)) AND 15
                  WordMovement = WordMovement + 1
               NEXT XWidth
            ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 1 THEN
               FOR XWidth = 0 TO TIFFInfo.TIFFWidth - 1
                  PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(MID$(BytesToExtractPerStrip$, WordMovement + 1, 1))
                  WordMovement = WordMovement + 1
               NEXT XWidth
            ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
               FOR XWidth = 0 TO TIFFInfo.TIFFWidth - 1
                  Red = ASC(MID$(BytesToExtractPerStrip$, WordMovement + 1, 1))
                  Green = ASC(MID$(BytesToExtractPerStrip$, WordMovement + 2, 1))
                  Blue = ASC(MID$(BytesToExtractPerStrip$, WordMovement + 3, 1))
                  PSET (XWidth * DividerX1!, YHeight * DividerY1!), (Red + Green + Blue) \ 3
                  WordMovement = WordMovement + 3
               NEXT XWidth
            END IF
            YHeight = YHeight + 1
            IF YHeight = TIFFInfo.TIFFHeight THEN
               EXIT FOR
            END IF
         NEXT RowNo
      NEXT StripNo
   ELSEIF TIFFInfo.TIFFCompression < 0 THEN
      YHeight = 0: Byte$ = SPACE$(1)
      IF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
         COLOR 240
         LOCATE 2: PRINT " Uncompressing 24-bit Image"
                   PRINT "  [Packbits Compression]"
         OPEN "24bit.tmp" FOR BINARY AS #2
      END IF
      FOR StripNo = 1 TO TIFFInfo.TIFFStrips
         SEEK #1, StripRows(StripNo, 1)
         FOR RowNo = 1 TO TIFFInfo.TIFFRowsinStrip
            XWidth = 0: XWidth1 = 0
            IF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
               LOCATE 1: PRINT YHeight; "/"; TIFFInfo.TIFFHeight; "Lines Completed"
            END IF
            DO
               GET #1, , Byte$
               Byte = ASC(Byte$)
               SELECT CASE Byte
                  CASE 0 TO 127:
                     ExtractBytes$ = SPACE$(Byte + 1)
                     GET #1, , ExtractBytes$
                     IF TIFFInfo.TIFFBitsPerPixel = 4 THEN
                        FOR Movement = 0 TO Byte
                           PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(MID$(ExtractBytes$, Movement + 1, 1)) \ 16
                           PSET ((XWidth * DividerX1!) + 1, YHeight * DividerY1!), ASC(MID$(ExtractBytes$, Movement + 1, 1)) AND 15
                           XWidth = XWidth + 2
                        NEXT Movement
                     ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 1 THEN
                        FOR Movement = 0 TO Byte
                           PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(MID$(ExtractBytes$, Movement + 1, 1))
                           XWidth = XWidth + 1
                        NEXT Movement
                     ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
                        PUT #2, , ExtractBytes$
                        XWidth1 = XWidth1 + Byte + 1
                     END IF
                  CASE 128
                  CASE ELSE
                     N = (ASC(Byte$) AND 127) - 128
                     N1 = (-1 * N) + 1
                     ExtractBytes$ = SPACE$(1)
                     GET #1, , ExtractBytes$
                     IF TIFFInfo.TIFFBitsPerPixel = 4 THEN
                        FOR Movement = 0 TO N1 - 1
                           PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(ExtractBytes$) \ 16
                           PSET ((XWidth * DividerX1!) + 1, YHeight * DividerY1!), ASC(ExtractBytes$) AND 15
                           XWidth = XWidth + 2
                        NEXT Movement
                     ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 1 THEN
                        FOR Movement = 0 TO N1 - 1
                           PSET (XWidth * DividerX1!, YHeight * DividerY1!), ASC(LEFT$(ExtractBytes$, 1))
                           XWidth = XWidth + 1
                        NEXT Movement
                     ELSEIF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
                        C$ = STRING$(N1, ASC(ExtractBytes$))
                        PUT #2, , C$
                        XWidth1 = XWidth1 + N1
                     END IF
               END SELECT
               IF XWidth = TIFFInfo.TIFFWidth OR XWidth1 = (TIFFInfo.TIFFWidth * 3) THEN
                  EXIT DO
               ELSEIF XWidth1 > (TIFFInfo.TIFFWidth * 3) THEN
                  LOCATE 4: PRINT " Error In TIFF File"
                  CLOSE : EXIT SUB
               END IF
            LOOP
            YHeight = YHeight + 1
            IF YHeight = TIFFInfo.TIFFHeight THEN
               EXIT FOR
            END IF
         NEXT RowNo
      NEXT StripNo
      IF TIFFInfo.TIFFBitsPerPixel = 8 AND TIFFInfo.TIFFPlanes = 3 THEN
         CLOSE #2
         OPEN "24bit.tmp" FOR BINARY AS #2
         Line$ = SPACE$(TIFFInfo.TIFFWidth * 3)
         FOR YHeight = 0 TO TIFFInfo.TIFFHeight
            GET #2, , Line$
            FOR XWidth = 0 TO TIFFInfo.TIFFWidth - 1
                Red = ASC(MID$(Line$, XWidth * 3 + 1, 1))
                Green = ASC(MID$(Line$, XWidth * 3 + 2, 1))
                Blue = ASC(MID$(Line$, XWidth * 3 + 3, 1))
                PSET (XWidth, YHeight), (Red + Green + Blue) \ 3
            NEXT XWidth
         NEXT YHeight
         CLOSE
         KILL "24bit.tmp"
      END IF
   END IF
ELSE
   PRINT " Another IFD was found"
   PRINT " Please Send File To <philipz85@hotmail.com>"
END IF
Wait$ = INPUT$(1)
CLOSE

END SUB

FUNCTION Sign2Unsign& (Integerx)

Sign2Unsign& = (Integerx AND 32768) + (Integerx AND 32767)

END FUNCTION

FUNCTION Unsign2Sign (Integerx&)

Unsign2Sign = (Integerx& AND 32767) - (Integerx& AND 32768)

END FUNCTION

